"""CLI.

CLI for converting D-Flow FM legacy external forcings files to current external forcings file/initial
fields file/structures file.
"""

import argparse
from argparse import ArgumentTypeError, Namespace
from pathlib import Path

from hydrolib.core import __version__
from hydrolib.core.base.utils import PathStyle
from hydrolib.tools.extforce_convert.main_converter import (
    ExternalForcingConverter,
    recursive_converter,
)


def valid_file(path_str: str):
    """Validate an .mdu file path and return it as a Path object.

    Args:
        path_str (str):
            The path to the MDU file as provided on the command line.

    Returns:
        Path: The validated path to the file.

    Raises:
        argparse.ArgumentTypeError: If the file does not exist or does not have a .mdu extension.
    """
    path = Path(path_str)
    if not str(path).lower().endswith(".mdu"):
        raise ArgumentTypeError(f"File must have a .mdu extension: {path}")

    if not path.exists():
        raise ArgumentTypeError(f"File not found: {path}")
    return path


def _validator(path_str: str, extension: str):
    """Validate that a file exists and matches the expected extension.

    Args:
        path_str (str):
            The path to the file to validate.
        extension (str):
            The required file extension (including the dot), e.g. ".ext".

    Returns:
        Path: The validated file path.

    Raises:
        argparse.ArgumentTypeError: If the file does not exist or the extension does not match.
    """
    path = Path(path_str)
    if not path.exists():
        raise ArgumentTypeError(f"File not found: {path}")
    if not str(path).lower().endswith(extension):
        raise ArgumentTypeError(f"File must have a {extension} extension: {path}")
    return path


def valid_file_with_extension(extension: str):
    """Create a validator callable for argparse that enforces a file extension.

    Args:
        extension (str):
            The required file extension (including the dot), e.g. ".mdu".

    Returns:
        Callable[[str], Path]:
            A function that validates a path string and returns a Path when invoked.

    Raises:
        argparse.ArgumentTypeError: Raised by the returned validator if the provided path does not exist or
            does not end with the required extension.
    """
    return lambda path_str: _validator(path_str, extension)


def valid_dir(path_str: str):
    """Validate that the given path exists and is a directory.

    Args:
        path_str (str): The path to validate.

    Returns:
        Path: The validated directory path.

    Raises:
        argparse.ArgumentTypeError: If the path does not exist or is not a directory.
    """
    path = Path(path_str)
    if not path.exists():
        raise ArgumentTypeError(f"Directory not found: {path}")
    if not path.is_dir():
        raise ArgumentTypeError(f"Path is not a directory: {path}")
    return path


def _get_parser() -> argparse.ArgumentParser:
    """Create and configure the argument parser for the extforce_convert CLI.

    Returns:
        argparse.ArgumentParser: The configured argument parser.
    """
    parser = argparse.ArgumentParser(
        prog="extforce_convert",
        description="Convert D-Flow FM legacy external forcings files to current external forcings file/initial fields file/structures file.",
    )

    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Print diagnostic information"
    )
    parser.add_argument(
        "--debug-mode",
        action="store_true",
        default=False,
        help="Convert the supported quantities only and leave unsupported quantities in the old external forcing "
        "file, default is False.(the conversion will fail if there is any unsupported quantities) ",
    )

    # mdu file, extforcefile and dir are mutually exclusive (can only use one)
    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "--mdufile",
        "-m",
        action="store",
        type=valid_file_with_extension(".mdu"),
        metavar="MDUFILE",
        help="Automatically take input and output filenames from MDUFILE.",
    )
    group.add_argument(
        "--extoldfile",
        "-e",
        action="store",
        type=valid_file_with_extension(".ext"),
        metavar="EXTOLDFILE",
        help="Input EXTOLDFILE to be converted.",
    )
    group.add_argument(
        "--dir",
        "-d",
        action="store",
        type=valid_dir,
        metavar="DIR",
        help="Directory to recursively find and convert .mdu files in.",
    )

    parser.add_argument(
        "--outfiles",
        "-o",
        action="store",
        nargs=3,
        metavar=("EXTFILE", "INIFIELDFILE", "STRUCTUREFILE"),
        help="Save forcings, initial fields and structures to specified filenames (only valid with --mdufile or --extoldfile).",
    )

    parser.add_argument(
        "--no-backup",
        dest="backup",
        action="store_false",
        help="Do not create a backup of each file that will be overwritten.",
    )
    parser.set_defaults(backup=True)

    parser.add_argument(
        "--remove-legacy-files",
        "-r",
        dest="remove_legacy",
        action="store_true",
        default=False,
        help="Remove legacy/old files (e.g. .tim) after conversion. Defaults to False.",
    )
    parser.add_argument(
        "--path-style",
        choices=[style.value for style in PathStyle],
        dest="path_style",
        default=None,
        type=lambda s: PathStyle(s),
        help="Handle absolute paths in input files according to the specified style (unix/windows)."
        "Use this when converting models with unix paths on Windows or windows paths on Unix.",
    )
    return parser


def main(args=None):
    """
    Entry point for the extforce_convert command-line tool.

    CLI argument combinations:

    Required (mutually exclusive, pick one):
      --mdufile, -m MDUFILE         Use MDUFILE to determine input/output files automatically.
      --extoldfile, -e EXTOLDFILE   Convert a specific legacy external forcing file.
      --dir, -d DIR                 Recursively find and convert all .mdu files in DIR.

    Optional:
      --outfiles, -o EXTFILE INIFIELDFILE STRUCTUREFILE
                               Specify output filenames for forcings, initial fields, and structures (only with --mdufile or --extoldfile).
                               Note: requires exactly three paths and is only valid for single-file conversions
                               (i.e., when using --mdufile or --extoldfile). Using --outfiles with --dir is invalid
                               and will result in an error.
      --no-backup                  Do not create a backup of overwritten files.
      --remove-legacy-files, -r    Remove legacy/old files (e.g. .tim) after conversion.
      --debug-mode                 Convert only supported quantities; leave unsupported quantities in the legacy external forcing file (default: False).
      --verbose, -v                Print diagnostic information.
      --version                    Print version and exit.
      --path-style {unix,windows}
                                 Handle absolute paths in input/output files according to the specified style (unix/windows).
                                 Use this when converting models that use unix paths but you are running the converter
                                 on a Windows machine, or the opposite.

    Args:
      args (Optional[List[str]]): Optional list of argument strings to parse instead of sys.argv. Useful for testing.

    Notes:
      - `--outfiles` cannot be combined with `--dir`.
      - `--outfiles` applies only to a single conversion target (from --mdufile or --extoldfile) and must provide three
        filenames, in this order: EXTFILE INIFIELDFILE STRUCTUREFILE.
      - When `--debug-mode` is provided, only supported quantities are converted; unsupported quantities remain in the
        legacy external forcing file. Without this flag, encountering unsupported quantities results in a failure.

    Examples (valid):
        - Use an MDU file and let the tool determine I/O automatically
            ```shell
            >>> extforce_convert --mdufile model.mdu # doctest: +SKIP
            ```
        - Convert a specific legacy .ext and explicitly set output filenames
            ```shell
            >>> extforce_convert --extoldfile old.ext --outfiles new.ext new.ini new.str # doctest: +SKIP
            ```
        - Recursively convert all models in a directory (no --outfiles here)
            ```shell
            >>> extforce_convert --dir ./models --no-backup --remove-legacy-files # doctest: +SKIP
            ```
        - Convert with explicit path style handling
            ```shell
            >>> extforce_convert --mdufile model.mdu --path-style unix # doctest: +SKIP
            ```
        - invalid flags combinations that will raise an error:
            --outfiles only works with single-file modes, not with --dir
            ```shell
            >>> extforce_convert --dir ./models --outfiles a.ext b.ini c.str # doctest: +SKIP
            ```
    """
    parser = _get_parser()
    args = parser.parse_args(args)

    # Disallow --outfiles when converting a directory.
    if args.dir is not None and args.outfiles is not None:
        parser.error(
            "--outfiles cannot be used with --dir. It only applies to single-file conversions."
        )

    if args.mdufile:
        convert_with_mdu_file(args)
    elif args.extoldfile is not None:
        convert_with_extold_file(args)
    elif args.dir is not None:
        recursive_converter(
            args.dir,
            backup=args.backup,
            remove_legacy=args.remove_legacy,
            debug=args.debug_mode,
        )
    else:
        print("Error: no input specified. Use one of --mdufile, --extoldfile or --dir.")


def convert_with_mdu_file(args: Namespace):
    """Convert the old external forcing file using the mdu file.

    Read the old external forcing file path from the mdu file,
    and convert it to the new format files.

    Args:
        args : Namespace
            The arguments parsed from the command line.
    """
    converter = ExternalForcingConverter.from_mdu(
        args.mdufile,
        ext_file_user=(args.outfiles[0] if args.outfiles else None),
        inifield_file_user=(args.outfiles[1] if args.outfiles else None),
        structure_file_user=(args.outfiles[2] if args.outfiles else None),
        path_style=args.path_style,
        debug=args.debug_mode,
    )
    convert(converter, args)


def convert_with_extold_file(args: Namespace):
    """Convert the old external forcing file to the new format files.

    Args:
        args : Namespace
            The arguments parsed from the command line.
    """
    converter = ExternalForcingConverter(
        args.extoldfile,
        ext_file=(args.outfiles[0] if args.outfiles else None),
        inifield_file=(args.outfiles[1] if args.outfiles else None),
        structure_file=(args.outfiles[2] if args.outfiles else None),
        path_style=args.path_style,
        debug=args.debug_mode,
    )
    convert(converter, args)


def convert(converter: ExternalForcingConverter, args: Namespace):
    """Run the update method of ExternalForcingConverter and save the results.

    Args:
        converter : ExternalForcingConverter
            The converter object to convert the old external forcing file.
        args : Namespace
            The arguments parsed from the command line.
    """
    converter.verbose = args.verbose
    converter.update()
    print("Converting the old external forcing file to the new format files is done.")
    print("Saving the new files ...")
    converter.save(backup=args.backup, recursive=True)
    print("The new files are saved.")
    if args.remove_legacy:
        print("Cleaning legacy tim files ...")
        converter.clean()


if __name__ == "__main__":
    main()



from enum import Enum
from pathlib import Path
from typing import Callable, Dict, List, Optional

from pydantic import Field

from hydrolib.core.base.models import BaseModel, ModelSaveSettings, ParsableFileModel
from hydrolib.core.dflowfm.cmp.parser import CMPParser
from hydrolib.core.dflowfm.cmp.serializer import CMPSerializer


class HarmonicRecord(BaseModel):
    """Single cmp record, representing a harmonic component with amplitude and phase.

    Args:
        period (float): the period.
        amplitude (float): the amplitude.
        phase (float): the phase in degrees.

    Returns:
        HarmonicRecord: A new instance of the `HarmonicRecord` class.

    Raises:
        ValueError: If the period, amplitude or phase are not valid numbers.

    Examples:
        Create a `HarmonicRecord` object from a dictionary:
            ```python
            >>> data = {
            ...     "period": 0.0,
            ...     "amplitude": 1.0,
            ...     "phase": 2.0
            ... }
            >>> harmonic_record = HarmonicRecord(**data)
            >>> print(harmonic_record.period)
            0.0

            ```
    """

    period: float
    amplitude: float
    phase: float


class AstronomicName(Enum):
    A0 = "A0"
    SA = "SA"
    SSA = "SSA"
    MSM = "MSM"
    MM = "MM"
    MSF = "MSF"
    MS0 = "MS0"
    MF = "MF"
    KO0 = "KO0"
    MK0 = "MK0"
    SNU = "SNU"
    SN = "SN"
    MSTM = "MSTM"
    MFM = "MFM"
    TwoSM = "2SM"
    MSQM = "MSQM"
    MQM = "MQM"
    TwoSMSMN = "2SMN"
    TwoSMOK1 = "2OK1"
    TwoSMQ1 = "2Q1"
    NJ1 = "NJ1"
    SIGMA1 = "SIGMA1"
    MUK1 = "MUK1"
    NUJ1 = "NUJ1"
    Q1 = "Q1"
    NK1 = "NK1"
    RO1 = "RO1"
    NUK1 = "NUK1"
    O1 = "O1"
    TAU1 = "TAU1"
    MP1 = "MP1"
    M1B = "M1B"
    M1C = "M1C"
    M1A = "M1A"
    M1 = "M1"
    NO1 = "NO1"
    CHI1 = "CHI1"
    LP1 = "LP1"
    PI1 = "PI1"
    TK1 = "TK1"
    P1 = "P1"
    SK1 = "SK1"
    S1 = "S1"
    K1 = "K1"
    MO1 = "MO1"
    SP1 = "SP1"
    PSI1 = "PSI1"
    RP1 = "RP1"
    FI1 = "FI1"
    KP1 = "KP1"
    THETA1 = "THETA1"
    LABDAO1 = "LABDAO1"
    J1 = "J1"
    MQ1 = "MQ1"
    TwoSMPO1 = "2PO1"
    SO1 = "SO1"
    OO1 = "OO1"
    TwoSMKO1 = "2KO1"
    UPSILON1 = "UPSILON1"
    KQ1 = "KQ1"
    TwoSMMN2S2 = "2MN2S2"
    ThreeMKS2 = "3MKS2"
    TwoSMNS2 = "2NS2"
    ThreeMS2 = "3MS2"
    OQ2 = "OQ2"
    MNK2 = "MNK2"
    EPSILON2 = "EPSILON2"
    MNS2 = "MNS2"
    TwoSMML2S2 = "2ML2S2"
    MNUS2 = "MNUS2"
    MNK2S2 = "MNK2S2"
    TwoSMMS2K2 = "2MS2K2"
    O2 = "O2"
    NLK2 = "NLK2"
    TwoSMMK2 = "2MK2"
    TwoSMN2 = "2N2"
    MU2 = "MU2"
    TwoSMMS2 = "2MS2"
    SNK2 = "SNK2"
    NA2 = "NA2"
    N2 = "N2"
    KQ2 = "KQ2"
    NB2 = "NB2"
    NU2 = "NU2"
    ThreeMSN2 = "3MSN2"
    TwoSMKN2S2 = "2KN2S2"
    OP2 = "OP2"
    MSK2 = "MSK2"
    GAMMA2 = "GAMMA2"
    ALFA2 = "ALFA2"
    MPS2 = "MPS2"
    MA2 = "MA2"
    M2 = "M2"
    KO2 = "KO2"
    MSP2 = "MSP2"
    MB2 = "MB2"
    DELTA2 = "DELTA2"
    MKS2 = "MKS2"
    M2KS2 = "M2(KS)2"
    TwoSMSNMK2 = "2SN(MK)2"
    LABDA2 = "LABDA2"
    SNM2 = "SNM2"
    TwoSMMN2 = "2MN2"
    L2 = "L2"
    L2A = "L2A"
    L2B = "L2B"
    TwoSMSK2 = "2SK2"
    T2 = "T2"
    S2 = "S2"
    KP2 = "KP2"
    R2 = "R2"
    K2 = "K2"
    MSNU2 = "MSNU2"
    MSN2 = "MSN2"
    ZETA2 = "ZETA2"
    ETA2 = "ETA2"
    KJ2 = "KJ2"
    MKN2 = "MKN2"
    TwoSMKMSN2 = "2KM(SN)2"
    TwoSMSM2 = "2SM2"
    SKM2 = "SKM2"
    TwoSMMS2N2 = "2MS2N2"
    TwoSMSNU2 = "2SNU2"
    TwoSMSN2 = "2SN2"
    SKN2 = "SKN2"
    MQ3 = "MQ3"
    NO3 = "NO3"
    MO3 = "MO3"
    TwoSMMK3 = "2MK3"
    TwoSMMP3 = "2MP3"
    M3 = "M3"
    NK3 = "NK3"
    SO3 = "SO3"
    MP3 = "MP3"
    MK3 = "MK3"
    SP3 = "SP3"
    TwoSMMQ3 = "2MQ3"
    SK3 = "SK3"
    TwoSMSO3 = "2SO3"
    K3 = "K3"
    FourMS4 = "4MS4"
    TwoSMMNS4 = "2MNS4"
    ThreeMK4 = "3MK4"
    MNLK4 = "MNLK4"
    ThreeMS4 = "3MS4"
    MSNK4 = "MSNK4"
    MN4 = "MN4"
    MNU4 = "MNU4"
    TwoSMMLS4 = "2MLS4"
    TwoSMMSK4 = "2MSK4"
    M4 = "M4"
    TwoSMMKS4 = "2MKS4"
    SN4 = "SN4"
    TwoSMMN4 = "3MN4"
    TwoSMSMK4 = "2SMK4"
    MS4 = "MS4"
    MK4 = "MK4"
    TwoSMSNM4 = "2SNM4"
    TwoSMMSN4 = "2MSN4"
    SL4 = "SL4"
    S4 = "S4"
    SK4 = "SK4"
    TwoSMSMN4 = "2SMN4"
    ThreeSM4 = "3SM4"
    TwoSMSKM4 = "2SKM4"
    MNO5 = "MNO5"
    ThreeMK5 = "3MK5"
    ThreeMP5 = "3MP5"
    M5 = "M5"
    MNK5 = "MNK5"
    TwoSMMP5 = "2MP5"
    MSO5 = "MSO5"
    ThreeMO5 = "3MO5"
    MSK5 = "MSK5"
    ThreeKM5 = "3KM5"
    TwoSMMNS6 = "2(MN)S6"
    ThreeMNS6 = "3MNS6"
    FourMK6 = "4MK6"
    TwoSMNM6 = "2NM6"
    FourMS6 = "4MS6"
    TwoSMMSNK6 = "2MSNK6"
    TwoSMMN6 = "2MN6"
    TwoSMMNU6 = "2MNU6"
    ThreeMSK6 = "3MSK6"
    M6 = "M6"
    MSN6 = "MSN6"
    MNK6 = "MNK6"
    FourMN6 = "4MN6"
    MKNU6 = "MKNU6"
    TwoSMMSK6 = "2(MS)K6"
    TwoSMMS6 = "2MS6"
    TwoSMMK6 = "2MK6"
    TwoSMSN6 = "2SN6"
    ThreeMSN6 = "3MSN6"
    MKL6 = "MKL6"
    TwoSMSM6 = "2SM6"
    MSK6 = "MSK6"
    S6 = "S6"
    TwoSMMNO7 = "2MNO7"
    TwoSMNMK7 = "2NMK7"
    M7 = "M7"
    TwoSMMSO7 = "2MSO7"
    MSKO7 = "MSKO7"
    TwoSMMN8 = "2(MN)8"
    ThreeMN8 = "3MN8"
    ThreeMNKS8 = "3MNKS8"
    M8 = "M8"
    TwoSMMSN8 = "2MSN8"
    TwoSMMNK8 = "2MNK8"
    ThreeMS8 = "3MS8"
    ThreeMK8 = "3MK8"
    TwoSMSNM8 = "2SNM8"
    MSNK8 = "MSNK8"
    TwoSMMS8 = "2(MS)8"
    TwoSMMSK8 = "2MSK8"
    ThreeSM8 = "3SM8"
    TwoSMSMK8 = "2SMK8"
    S8 = "S8"
    TwoSMMNK9 = "2(MN)K9"
    ThreeMNK9 = "3MNK9"
    FourMK9 = "4MK9"
    ThreeMSK9 = "3MSK9"
    FourMN10 = "4MN10"
    M10 = "M10"
    ThreeMSN10 = "3MSN10"
    FourMS10 = "4MS10"
    TwoSMMSN10 = "2(MS)N10"
    TwoSMMNSK10 = "2MNSK10"
    ThreeM2S10 = "3M2S10"
    FourMSK11 = "4MSK11"
    M12 = "M12"
    FourMSN12 = "4MSN12"
    FourMS12 = "5MS12"
    ThreeMNKS12 = "3MNKS12"
    FourM2S12 = "4M2S12"


class AstronomicRecord(BaseModel):
    """Single cmp record, representing an astronomic component with amplitude and phase.

    Args:
        name (AstronomicName): the astronomic name.
        amplitude (float): the amplitude.
        phase (float): the phase in degrees.

    Returns:
        AstronomicRecord: A new instance of the `AstronomicRecord` class.

    Raises:
        ValueError: If the name, amplitude or phase are not valid numbers.

    Examples:
        Create an `AstronomicRecord` object from a dictionary:
            ```python
            >>> data = {
            ...     "name": "4MS10",
            ...     "amplitude": 1.0,
            ...     "phase": 2.0
            ... }
            >>> astronomic_record = AstronomicRecord(**data)
            >>> print(astronomic_record.name)
            4MS10

            ```
    """

    name: AstronomicName
    amplitude: float
    phase: float


class CMPSet(BaseModel):
    """A CMP set containing harmonics and astronomics.

    Args:
        harmonics (List[HarmonicRecord]): A list containing the harmonic components.
        astronomics (List[AstronomicRecord]): A list containing the astronomic components.

    Returns:
        CmpSet: A new instance of the `CmpSet` class.

    Raises:
        ValueError: If the harmonics or astronomics are not valid lists.

    Examples:
        Create a `CmpSet` object from a dictionary:
            ```python
            >>> data = {
            ...     "astronomics": [{"name": "4MS10", "amplitude": 1.0, "phase": 2.0}]
            ... }
            >>> cmp_set = CMPSet(**data)
            >>> print(cmp_set.astronomics)
            [AstronomicRecord(name='4MS10', amplitude=1.0, phase=2.0)]

            ```
    """

    harmonics: Optional[List[HarmonicRecord]] = Field(default_factory=list)
    astronomics: Optional[List[AstronomicRecord]] = Field(default_factory=list)


class CMPModel(ParsableFileModel):
    """Class representing a cmp (*.cmp) file.
    This class is used to parse and serialize cmp files, which contain
    information about various components such as harmonics and astronomics.

    Args:
        comments (List[str]): A list with the header comment of the cmp file.
        components (CMPSet): A CMPSet of astronomic and/or harmonic records.

    Returns:
        CmpModel: A new instance of the `CmpModel` class.

    Raises:
        ValueError: If the comments or components are not valid lists.

    Examples:
        Create a `CmpModel` object from a dictionary:
            ```python
            >>> data = {
            ...     "comments": ["# Example comment"],
            ...     "component": {
            ...         "harmonics": [{"period": 0.0, "amplitude": 1.0, "phase": 2.0}],
            ...         "astronomics": [{"name": "4MS10", "amplitude": 1.0, "phase": 2.0}]
            ...     }
            ... }
            >>> cmp_model = CMPModel(**data)
            >>> print(cmp_model.component.astronomics)
            [AstronomicRecord(name='4MS10', amplitude=1.0, phase=2.0)]

            ```

    See Also:
        CmpSet: Class representing the components of the cmp file.
        CmpSerializer: Class responsible for serializing cmp files.
        CmpParser: Class responsible for parsing cmp files.
    """

    comments: List[str] = Field(default_factory=list)
    component: CMPSet = Field(default_factory=CMPSet)
    quantities_name: Optional[List[str]] = None

    @classmethod
    def _ext(cls) -> str:
        return ".cmp"

    @classmethod
    def _filename(cls) -> str:
        return "components"

    @classmethod
    def _get_serializer(
        cls,
    ) -> Callable[[Path, Dict, ModelSaveSettings], None]:
        return CMPSerializer.serialize

    @classmethod
    def _get_parser(cls) -> Callable[[Path], Dict]:
        return CMPParser.parse

    def get_units(self):
        """Return the units for each quantity in the timeseries.

        Returns:
            List[str]: A list of units for each quantity in the timeseries.

        Examples:
            Create a `CMPModel` object from a .cmp file:
                ```python
                >>> data = {
                ...     "comments": ["# Example comment"],
                ...     "component": {
                ...         "harmonics": [{"period": 0.0, "amplitude": 1.0, "phase": 2.0}],
                ...     },
                ...     "quantities_name": ["discharge"],
                ... }
                >>> model = CMPModel(**data)
                >>> print(model.get_units())
                ['m3/s']

                ```
        """
        if self.quantities_name is None:
            return None
        return CMPModel._get_quantity_unit(self.quantities_name)



from pathlib import Path
from typing import Dict


class DummySerializer:
    @staticmethod
    def serialize(path: Path, data: Dict, config, save_settings) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w", encoding="utf8") as f:
            f.write(str(data))



from pathlib import Path
from typing import Any, Dict, List, Tuple

from hydrolib.core.base.parser import BaseParser

AstronomicData = Dict[str, Tuple[float, float, float]]
HarmonicData = Dict[str, Tuple[str, float, float]]
CMPData = Dict[str, Tuple[AstronomicData, HarmonicData]]


class CMPParser(BaseParser):
    """Parser for .cmp files.
    Full line comments at the start of the file are supported. Comment lines start with either a `*` or a `#`.
    No other comments are supported.
    """

    @staticmethod
    def parse(filepath: Path) -> Dict[str, List[Any]]:
        r"""Parse a cmp file to a Dict containing amplitude and phase for period.

        Args:
            filepath (Path): Path to the .cmp file to be parsed.

        Returns:
            Dict[str, List[Any]]: A dictionary with keys "comments" and "components".
            - "comments" represents comments found at the start of the file.
            - "components" is a list of dictionaries with the values "period", "amplitude" and "phase".
                - "period" is a float as a string.
                - "amplitude" is a float as a string.
                - "phase" is a float as a string.

        Raises:
            ValueError: If the file contains a comment that is not at the start of the file.

        Examples:
            Read a `.cmp` file:
                ```python
                >>> file_path = Path(
                ...    "tests/data/input/e02/f006_external_forcing/c020_basinnofriction_squares/openboundary1_0001.cmp"
                ... )
                >>> cmp_model = CMPParser.parse(file_path)
                >>> print(cmp_model)  # doctest: +ELLIPSIS
                {'comments': [' COLUMNN=3', ' COLUMN1=Period (min) or Astronomical Componentname',... 'amplitude': '1.0000000', 'phase': '45.1200000'}]}}

                ```
        """
        with filepath.open(encoding="utf8") as file:
            lines = file.readlines()
            comments, start_components_index = CMPParser._read_header_comments(lines)
            component = CMPParser._read_components_data(lines, start_components_index)
        return {"comments": comments, "component": component}

    @staticmethod
    def _read_components_data(
        lines: List[str], start_components_index: int
    ) -> Dict[str, Any]:
        harmonics_data: List[Dict[str, Any]] = []
        astronomics_data: List[Dict[str, Any]] = []
        for line_index in range(start_components_index, len(lines)):
            line = lines[line_index].strip()

            if len(line) == 0:
                continue

            CMPParser._raise_error_if_contains_comment(line, line_index + 1)

            period, amplitude, phase = line.split()

            if CMPParser._is_float(period):
                component = {"period": period, "amplitude": amplitude, "phase": phase}
                harmonics_data.append(component)
            else:
                component = {"name": period, "amplitude": amplitude, "phase": phase}
                astronomics_data.append(component)

        component_data = {}
        if harmonics_data:
            component_data["harmonics"] = harmonics_data
        if astronomics_data:
            component_data["astronomics"] = astronomics_data

        return component_data

    @staticmethod
    def _is_float(element: any) -> bool:
        if element is None:
            return False
        try:
            float(element)
            return True
        except ValueError:
            return False



from pathlib import Path
from typing import Any, Dict, List

from hydrolib.core.base.parser import BaseParser

TimData = Dict[str, List[str]]


class TimParser(BaseParser):
    """
    A parser for .tim files.
    Full line comments at the start of the file are supported. Comment lines start with either a `*` or a `#`.
    No other comments are supported.
    """

    @staticmethod
    def parse(filepath: Path) -> Dict[str, List[Any]]:
        """Parse a .tim file into a dictionary with comments and time series data.

        Args:
            filepath (Path): Path to the .tim file to be parsed.

        Returns:
            Dict[str, List[Any]]: A dictionary with keys "comments" and "timeseries".
            - "comments" represents comments found at the start of the file.
            - "timeseries" is a list of dictionaries with the key as "time" and values as "data".
                - "time" is a time as a string.
                - "data" is data as a list of strings.

        Raises:
            ValueError: If the file contains a comment that is not at the start of the file.
            ValueError: If the data of the timeseries is empty.
        """
        with filepath.open(encoding="utf8") as file:
            lines = file.readlines()
            comments, start_timeseries_index = TimParser._read_header_comments(lines)
            timeseries = TimParser._read_time_series_data(lines, start_timeseries_index)

        return {"comments": comments, "timeseries": timeseries}

    @staticmethod
    def _read_time_series_data(
        lines: List[str], start_timeseries_index: int
    ) -> List[TimData]:
        timeseries: List[TimData] = []
        for line_index in range(start_timeseries_index, len(lines)):
            line = lines[line_index].strip()

            if len(line) == 0:
                continue

            TimParser._raise_error_if_contains_comment(line, line_index + 1)

            time, *values = line.split()

            TimParser._raise_error_if_values_empty(values, line_index)

            timrecord = {"time": time, "data": values}
            timeseries.append(timrecord)

        return timeseries

    @staticmethod
    def _raise_error_if_values_empty(values: List[str], line_index: int) -> None:
        if len(values) == 0:
            raise ValueError(f"Line {line_index}: Time series cannot be empty.")



"""Utility functions for converting old external forcing files to new format."""

import os
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, Iterable, List, Set, Type, Union

import yaml
from pydantic import BaseModel, ConfigDict, Extra, Field, field_validator

from hydrolib import __path__
from hydrolib.core.base.file_manager import PathOrStr
from hydrolib.core.base.models import DiskOnlyFileModel, FileModel
from hydrolib.core.dflowfm.ext.models import (
    MeteoForcingFileType,
    MeteoInterpolationMethod,
)
from hydrolib.core.dflowfm.extold.models import (
    ExtOldFileType,
    ExtOldForcing,
    ExtOldModel,
    ExtOldQuantity,
)
from hydrolib.core.dflowfm.inifield.models import (
    AveragingType,
    DataFileType,
    InterpolationMethod,
)

SOURCESINK_SALINITY_IN_BC = "sourcesink_salinitydelta"
SOURCESINK_TEMP_IN_BC = "sourcesink_temperaturedelta"
SOURCESINK_NAME_IN_EXT = "discharge_salinity_temperature_sorsin"


__all__ = [
    "UnSupportedQuantitiesError",
    "CONVERTER_DATA",
    "find_temperature_salinity_in_quantities",
    "IgnoreUnknownKeyWordClass",
    "backup_file",
    "construct_filemodel_new_or_existing",
    "path_relative_to_parent",
]


AVERAGING_TYPE_DICT = {
    1: AveragingType.mean,
    2: AveragingType.nearestnb,
    3: AveragingType.max,
    4: AveragingType.min,
    5: AveragingType.invdist,
    6: AveragingType.minabs,
    7: AveragingType.median,
}


CONVERTER_DATA_PATH = Path(__path__[0]) / "tools/extforce_convert/data/data.yaml"
with CONVERTER_DATA_PATH.open("r") as fh:
    try:
        CONVERTER_DATA = yaml.safe_load(fh)
    except yaml.YAMLError as e:
        raise RuntimeError(f"Failed to parse YAML at {CONVERTER_DATA_PATH}: {e}") from e


def construct_filemodel_new_or_existing(
    model_class: Type[FileModel], filepath: PathOrStr, *args, **kwargs
) -> FileModel:
    """Construct a new instance of a FileModel subclass, either loading an existing model file, or starting a blank one.

    If the given filepath exists, it will be loaded, as if the constructor
    was directly called. If it does not exist, the blank model instance
    will only have its filepath attribute set, e.g., for future saving.

    Args:
        model_class (type[FileModel]): The FileModel subclass to construct or update.
        filepath (PathOrStr): The filepath to use for the new or existing model input.
        *args: Additional positional arguments to pass to the ModelClass constructor
        **kwargs: Additional keywords arguments to pass to the ModelClass constructor
    """
    if Path(filepath).is_file():
        model = model_class(filepath=filepath, *args, **kwargs)
    else:
        model = model_class(*args, **kwargs)
        model.filepath = filepath

    return model


def backup_file(filepath: PathOrStr) -> None:
    """Create a backup of the given file by copying it to a new file with a '.bak' extension.

    Args:
        filepath (PathOrStr): The path to the file to back up.
    """
    filepath = Path(filepath) if isinstance(filepath, str) else filepath
    if filepath.is_file():
        backup_path = filepath.with_suffix(filepath.suffix + ".bak")
        filepath.replace(backup_path)


def oldfiletype_to_forcing_file_type(
    oldfiletype: int,
) -> Union[MeteoForcingFileType, str]:
    """Convert old external forcing `FILETYPE` integer value to valid `forcingFileType` string value.

    Args:
        oldfiletype (int): The FILETYPE value in an old external forcings file.

    Returns:
        Union[MeteoForcingFileType,str]: Corresponding value for `forcingFileType`,
            or "unknown" for invalid input.
    """
    forcing_file_type = "unknown"

    if oldfiletype == ExtOldFileType.TimeSeries:  # 1
        forcing_file_type = MeteoForcingFileType.uniform
    elif oldfiletype == ExtOldFileType.TimeSeriesMagnitudeAndDirection:  # 2
        forcing_file_type = MeteoForcingFileType.unimagdir
    elif oldfiletype == ExtOldFileType.SpatiallyVaryingWindPressure:  # 3
        raise NotImplementedError(
            "FILETYPE = 3 (spatially verying wind and pressure) is no longer supported."
        )
    elif oldfiletype == ExtOldFileType.ArcInfo:  # 4
        forcing_file_type = MeteoForcingFileType.arcinfo
    elif oldfiletype == ExtOldFileType.SpiderWebData:  # 5
        forcing_file_type = MeteoForcingFileType.spiderweb
    elif oldfiletype == ExtOldFileType.CurvilinearData:  # 6
        forcing_file_type = MeteoForcingFileType.curvigrid
    elif oldfiletype == ExtOldFileType.Samples:  # 7
        forcing_file_type = DataFileType.sample
    elif oldfiletype == ExtOldFileType.TriangulationMagnitudeAndDirection:  # 8
        raise NotImplementedError(
            "FILETYPE = 8 (magnitude+direction timeseries on stations) is no longer supported."
        )
    elif oldfiletype == ExtOldFileType.Polyline:  # 9
        # Boundary polyline files no longer need a filetype of their own (intentionally no error raised)
        pass
    elif oldfiletype == ExtOldFileType.InsidePolygon:  # 10
        forcing_file_type = DataFileType.polygon
    elif oldfiletype == ExtOldFileType.NetCDFGridData:  # 11
        forcing_file_type = MeteoForcingFileType.netcdf

    return forcing_file_type


def oldmethod_to_interpolation_method(
    oldmethod: int,
) -> Union[InterpolationMethod, MeteoInterpolationMethod, str]:
    """Convert old external forcing `METHOD` integer value to valid `interpolationMethod` string value.

    Args:
        oldmethod (int): The METHOD value in an old external forcings file.

    Returns:
        Union[InterpolationMethod,str]: Corresponding value for `interpolationMethod`,
            or "unknown" for invalid input.
    """
    if oldmethod in [1, 2, 3, 11]:
        interpolation_method = MeteoInterpolationMethod.linearSpaceTime
    elif oldmethod == 5:
        interpolation_method = InterpolationMethod.triangulation
    elif oldmethod == 4:
        interpolation_method = InterpolationMethod.constant
    elif oldmethod in range(6, 10):
        interpolation_method = InterpolationMethod.averaging
    else:
        interpolation_method = "unknown"
    return interpolation_method


def map_method_to_averaging_type(
    old_forcing_method: int,
    averaging_type: int,
) -> Union[AveragingType, str]:
    """Convert an old external forcing `METHOD` integer value to a valid ` averagingType ` string value.

    Args:
        old_forcing_method (int):
            The `METHOD` value in an old external forcings file.
        averaging_type (int):
            The `AVERAGINGTYPE` value in an old external forcings file.
            AVERAGINGTYPE (ONLY WHEN METHOD=6)
            ```ini
            =1  : SIMPLE AVERAGING
            =2  : NEAREST NEIGHBOUR
            =3  : MAX (HIGHEST)
            =4  : MIN (LOWEST)
            =5  : INVERSE WEIGHTED DISTANCE-AVERAGE
            =6  : MINABS
            =7  : KDTREE (LIKE 1, BUT FAST AVERAGING)
            ```

    Notes:
        - The new external forcing will have an Averaging type if the old external forcings had a `Method = 6`.


    Returns:
        Union[AveragingType,str]:
            Corresponding value for `averagingType`, or "unknown" for invalid input.
    """
    if old_forcing_method == 6:
        if averaging_type is None:
            averaging_type = AveragingType.mean
        else:
            averaging_type = AVERAGING_TYPE_DICT.get(int(averaging_type), "unknown")
    else:
        averaging_type = "unknown"

    return averaging_type


def convert_interpolation_data(
    forcing: ExtOldForcing, data: Dict[str, Any]
) -> Dict[str, str]:
    """Convert interpolation data from old to new format.

    Args:
        forcing (ExtOldForcing): The old forcing block with interpolation data.
        data (Dict[str, Any]): The dictionary to which the new data will be added.

    Returns:
        Dict[str, str]: The updated dictionary with the new interpolation data.
        - The dictionary will contain the "interpolationmethod" key with the new interpolation method.
        - if the interpolation method is "Averaging" (method = 6), the dictionary will also contain
            the "averagingtype", "averagingrelsize", "averagingnummin", and "averagingpercentile" keys.
    """
    data["interpolationmethod"] = oldmethod_to_interpolation_method(forcing.method)
    if data["interpolationmethod"] == InterpolationMethod.averaging:
        data["averagingtype"] = map_method_to_averaging_type(
            forcing.method, forcing.averagingtype
        )
        data["averagingrelsize"] = forcing.relativesearchcellsize
        data["averagingnummin"] = forcing.nummin
        data["averagingpercentile"] = forcing.percentileminmax

    return data


def path_relative_to_parent(
    forcing: ExtOldForcing,
    inifile_path: Path,
    ext_old_path: Path,
    mdu_parser: Any,
) -> Path:
    """Resolve the path of the forcing file relative to the parent file if needed.

    Args:
        forcing (ExtOldForcing):
            The old forcing block with the filename to resolve.
        inifile_path (Path):
            The path to the inifields file.
        ext_old_path (Path):
            The path to the old external forcing file.
        mdu_parser (MDUParser):
            The MDU parser object containing the loaded MDU data.
            This holds the "pathsRelativeToParent" setting.

    Returns:
        Path: The resolved path of the forcing file.
    """
    if mdu_parser is None:
        resolve_parent = False
    else:
        resolve_parent = mdu_parser.is_relative_to_parent

    update_path_condition = (
        forcing.filename.filepath.is_absolute() or not resolve_parent
    )

    forcing_path = (
        forcing.filename.filepath
        if update_path_condition
        else os.path.relpath(
            ext_old_path.parent / forcing.filename.filepath, inifile_path.parent
        )
    )
    return forcing_path


def create_initial_cond_and_parameter_input_dict(
    forcing: ExtOldForcing,
    new_forcing_path: Path,
) -> Dict[str, str]:
    """Create the input dictionary for the `InitialField` or `ParameterField`.

    Args:
        forcing: [ExtOldForcing]
            External forcing block from the old external forcings file.
        new_forcing_path: [Path]
            The path to the new forcing file.

    Returns:
        Dict[str, str]:
            the input dictionary to the `InitialField` or `ParameterField` constructor
    """
    quantity_name = (
        forcing.quantity
        if forcing.quantity != ExtOldQuantity.BedRockSurfaceElevation
        else "bedrockSurfaceElevation"
    )
    block_data = {
        "quantity": quantity_name,
        "datafile": DiskOnlyFileModel(new_forcing_path),
        "datafiletype": oldfiletype_to_forcing_file_type(forcing.filetype),
    }
    if block_data["datafiletype"] == "polygon":
        block_data["value"] = forcing.value

    if forcing.sourcemask != DiskOnlyFileModel(None):
        raise ValueError(
            f"Attribute 'SOURCEMASK' is no longer supported, cannot "
            f"convert this input. Encountered for QUANTITY="
            f"{forcing.quantity} and FILENAME={forcing.filename}."
        )
    block_data = convert_interpolation_data(forcing, block_data)
    block_data["operand"] = forcing.operand

    if hasattr(forcing, "extrapolation"):
        block_data["extrapolationmethod"] = (
            "yes" if forcing.extrapolation == 1 else "no"
        )
    for key, value in forcing.model_dump().items():
        if key.lower().startswith("tracer") and value is not None:
            block_data[key] = value
    return block_data


def find_temperature_salinity_in_quantities(strings: List[str]) -> Dict[str, int]:
    """Find temperature and salinity in quantities.

    Searches for keywords "temperature" and "salinity" in a list of strings
    and returns a dictionary with associated values.

    Args:
        strings (List[str]):
            A list of strings to search.

    Returns:
        Dict[str, int]:
            A dictionary with keys as "salinity" or "temperature" and values 3 and 4 respectively.

     Examples:
        ```python
        >>> from hydrolib.tools.extforce_convert.utils import find_temperature_salinity_in_quantities
        >>> find_temperature_salinity_in_quantities(["temperature", "Salinity"])
        OrderedDict({'sourcesink_salinitydelta': 3, 'sourcesink_temperaturedelta': 4})
        >>> find_temperature_salinity_in_quantities(["Temperature"])
        OrderedDict({'sourcesink_temperaturedelta': 3})
        >>> find_temperature_salinity_in_quantities(["Salinity"])
        OrderedDict({'sourcesink_salinitydelta': 3})
        >>> find_temperature_salinity_in_quantities(["tracers"])
        OrderedDict()
        >>> find_temperature_salinity_in_quantities([])
        OrderedDict()
        >>> find_temperature_salinity_in_quantities(["discharge_salinity_temperature_sorsin", "Salinity"])
        OrderedDict({'sourcesink_salinitydelta': 3})

        ```

    Notes:
        - The function removes the `discharge_salinity_temperature_sorsin` from the given list of strings.
        - The function removes the duplicate strings in the list.
    """
    result = OrderedDict()
    strings = list(set(strings))
    # remove the `discharge_salinity_temperature_sorsin` quantity from the list
    if SOURCESINK_NAME_IN_EXT in strings:
        strings.remove(SOURCESINK_NAME_IN_EXT)

    if any("salinity" in string.lower() for string in strings):
        result[SOURCESINK_SALINITY_IN_BC] = 3
    if any("temperature" in string.lower() for string in strings):
        result[SOURCESINK_TEMP_IN_BC] = (
            result.get(SOURCESINK_SALINITY_IN_BC, 2) + 1
        )  # Default temperature value is 2

    return result


class IgnoreUnknownKeyWord(type):
    """Metaclass to ignore unknown keyword arguments when creating a new class instance."""

    def __call__(cls, base_class, **data):
        """Dynamically create and instantiate a subclass of base_class."""

        class DynamicClass(base_class):
            model_config = ConfigDict(extra="ignore")

            def __init__(self, **data):
                valid_fields = self.__annotations__.keys()
                filtered_data = {k: v for k, v in data.items() if k in valid_fields}
                super().__init__(**filtered_data)

        return DynamicClass(**data)


class IgnoreUnknownKeyWordClass(metaclass=IgnoreUnknownKeyWord):
    """Base class to ignore unknown keyword arguments when creating a new class instance."""

    pass


def check_unique(v):
    """Checks and filters unique non-empty strings from the input list.

        This function processes a given input list, ensuring that all strings are:
        - Trimmed of any leading or trailing whitespace.
        - Converted to lowercase for case-insensitive uniqueness.
        - Added to the result list only if they haven't been seen before.

        Empty strings or non-string elements are ignored.

    Args:
        v (list[str] | None):
            A list of strings to be processed. May contain None or non-string elements.
            If the input is None, it is treated as an empty list.

    Returns:
        list[str]
            A list of unique, lowercase, trimmed strings in the order of their first
            appearance.
    """
    seen = set()
    unique = []
    for s in v or []:
        if isinstance(s, str):
            key = s.strip().lower()
            if key and key not in seen:
                seen.add(key)
                unique.append(key)
    return unique


class MDUConfig(BaseModel):
    deprecated_keywords: Set[str] = Field(default_factory=set)
    deprecated_value: int = 0

    @field_validator("deprecated_keywords", mode="before")
    def _to_set(cls, v):
        """convert the deprecated keywords to a set."""
        if v is None:
            return set()

        if isinstance(v, str):
            v = [v]
        return set(check_unique(v))


class ExternalForcingConfigs(BaseModel):
    unsupported_quantity_names: List[str] = Field(default_factory=list)
    unsupported_prefixes: List[str] = Field(default_factory=list)

    @field_validator(
        "unsupported_quantity_names", "unsupported_prefixes", mode="before"
    )
    def ensure_unique(cls, v: List[str]) -> List[str]:
        return check_unique(v)

    def find_unsupported(self, quantities: Iterable[str]) -> Set[str]:
        """Return the set of unsupported quantities present in the given iterable."""
        normalized = [str(q).lower() for q in quantities]
        result = set(self.unsupported_quantity_names).intersection(normalized)

        for q in normalized:
            if any(q.startswith(p) for p in self.unsupported_prefixes):
                result.add(q)
        return result

    def check_unsupported_quantities(
        self, quantities: Iterable[str], raise_error: bool = True
    ) -> Set[str]:
        """Raise an error if any of the given quantities are unsupported."""
        un_supported = self.find_unsupported(quantities)
        if raise_error and un_supported:
            raise UnSupportedQuantitiesError(
                f"The following quantities are not supported by the converter yet: {un_supported}"
            )
        else:
            return un_supported


class ConverterData(BaseModel):
    version: float
    mdu: MDUConfig = Field(default_factory=MDUConfig)
    external_forcing: ExternalForcingConfigs = Field(
        default_factory=ExternalForcingConfigs
    )

    def check_unsupported_quantities(
        self, ext_old_model: ExtOldModel, raise_error: bool = True
    ):
        """Check if the old external forcing file contains unsupported quantities."""
        quantities = [forcing.quantity for forcing in ext_old_model.forcing]
        unsupported_quantities = self.external_forcing.check_unsupported_quantities(
            quantities, raise_error=raise_error
        )
        return unsupported_quantities


CONVERTER_DATA = ConverterData(**CONVERTER_DATA)


class UnSupportedQuantitiesError(Exception):
    pass



"""File Manager Module."""

from contextlib import contextmanager
from contextvars import ContextVar
from enum import IntEnum
from pathlib import Path
from typing import Dict, Generator, List, Optional, Tuple, Union

from hydrolib.core.base.utils import (
    FileChecksumCalculator,
    FilePathStyleConverter,
    OperatingSystem,
    PathStyle,
    get_operating_system,
    get_path_style_for_current_operating_system,
    str_is_empty_or_none,
)

PathOrStr = Union[Path, str]
# We use ContextVars to keep a reference to the folder
# we're currently parsing files in. In the future
# we could move to https://github.com/samuelcolvin/pydantic/issues/1549
context_file_loading: ContextVar["FileLoadContext"] = ContextVar("file_loading")


class ResolveRelativeMode(IntEnum):
    """ResolveRelativeMode.

    ResolveRelativeMode defines the possible resolve modes used within the
    FilePathResolver.

    It determines how the relative paths to child models within some FileModel
    should be interpreted. By default it should be relative to the parent model,
    i.e. the model in which the mode is defined. However there exist some exceptions,
    where the relative paths should be evaluated relative to some other folder,
    regardless of the current parent location.

    Options:
        ToParent:
            Relative paths should be resolved relative to their direct parent model.
        ToAnchor:
            All relative paths should be resolved relative to the specified regardless
            of subsequent parent model locations.
    """

    ToParent = 0
    ToAnchor = 1


class FilePathResolver:
    """FilePathResolver is responsible for resolving relative paths.

    The FilePathResolver maintains a stack of parent paths and their associated resolve modes.
    It provides functionality to resolve relative paths to absolute paths based on the current state.
    The current state to which paths are resolved can be altered by pushing a new parent path
    to the FilePathResolver, or removing the latest added parent path from the FilePathResolver.

    The resolver supports two modes:
    1. ToParent: Resolves paths relative to the direct parent
    2. ToAnchor: Resolves paths relative to a specified anchor path

    Examples:
        - Mode ToParent:
        ```python
        >>> from hydrolib.core.base.file_manager import FilePathResolver, ResolveRelativeMode
        >>> from pathlib import Path
        >>> resolver = FilePathResolver()
        >>> parent_path = Path("C:/project")
        >>> resolver.push_new_parent(parent_path, ResolveRelativeMode.ToParent)
        >>> print(resolver._parents) # doctest: +SKIP
        [(WindowsPath('C:/project'), <ResolveRelativeMode.ToParent: 0>)]

        ```
        - Now the resolver has one parent path "C:/project" and will resolve paths relative to it.

        ```python
        >>> my_file = Path("data/file.txt")
        >>> resolver.resolve(my_file) # doctest: +SKIP
        Path('C:/project/data/file.txt')

        ```
        - adding new parent paths will change the current parent path:
        ```python
        >>> new_parent = Path("models")
        >>> resolver.push_new_parent(new_parent, ResolveRelativeMode.ToParent)
        >>> print(resolver._parents) # doctest: +SKIP
        [(WindowsPath('C:/project'), <ResolveRelativeMode.ToParent: 0>), (WindowsPath('C:/project/models'), <ResolveRelativeMode.ToParent: 0>)]
        ```
        - Now the resolver have two parents and it will resolve paths relative to the new parent path:
        ```python
        >>> my_file = Path("model1.txt")
        >>> resolver.resolve(my_file) # doctest: +SKIP
        Path('C:/project/models/model1.txt')
        ```
        - Mode ToAnchor:
        ```pyhon
        >>> resolver = FilePathResolver()
        >>> anchor_path = Path("C:/anchor")
        >>> resolver.push_new_parent(anchor_path, ResolveRelativeMode.ToAnchor)
        >>> print(resolver._anchors)
        [WindowsPath('C:/anchor')]
        >>> print(resolver._parents)
        [(WindowsPath('C:/anchor'), <ResolveRelativeMode.ToAnchor: 1>)]
        >>> new_parent = Path("subdir")
        >>> resolver.push_new_parent(new_parent, ResolveRelativeMode.ToParent)
        >>> resolver.resolve(Path("file.txt")) # doctest: +SKIP
        Path('C:/anchor/file.txt')

        ```
    """

    def __init__(self) -> None:
        """Initialize a new empty FilePathResolver.

        Creates a new FilePathResolver with empty stacks for anchors and parents.
        By default, relative paths will be resolved relative to the current working directory.

        Examples:
            ```python
            >>> resolver = FilePathResolver()
            >>> resolver.get_current_parent() #doctest: +SKIP
            Path('/current/working/directory')

            ```
        """
        self._anchors: List[Path] = []
        self._parents: List[Tuple[Path, ResolveRelativeMode]] = []

    @property
    def _anchor(self) -> Optional[Path]:
        """Get the latest anchor path if available.

        Returns:
            Optional[Path]:
                The latest anchor path if any anchors have been added, None otherwise.
        """
        return self._anchors[-1] if self._anchors else None

    @property
    def _direct_parent(self) -> Path:
        """Get the latest parent path or current working directory.

        Returns:
            Path:
                The latest parent path if any parents have been added, or the current working directory if no
                parents have been added.
        """
        return self._parents[-1][0] if self._parents else Path.cwd()

    def get_current_parent(self) -> Path:
        """Get the current absolute path with which files are resolved.

        Determines the current parent path based on the active resolve mode:
        - If the current mode is ResolveRelativeMode.ToAnchor, returns the latest added anchor path.
        - If the current mode is ResolveRelativeMode.ToParent or no mode is set, returns the latest added parent path.

        Returns:
            Path:
                The absolute path to the current parent.

        Examples:
            ```python
            >>> resolver = FilePathResolver()
            >>> resolver.push_new_parent(Path("C:/project"), ResolveRelativeMode.ToParent)
            >>> resolver.get_current_parent() # doctest: +SKIP
            Path('C:/project')

            >>> resolver.push_new_parent(Path("C:/anchor"), ResolveRelativeMode.ToAnchor)
            >>> resolver.get_current_parent() # doctest: +SKIP
            Path('C:/anchor')

            ```
        """
        if self._anchor:
            return self._anchor
        return self._direct_parent

    def resolve(self, path: Path) -> Path:
        """Resolve the provided path to an absolute path given the current state.

        If the provided path is already absolute, it will be returned as is.
        Otherwise, the path is resolved relative to the current parent path,
        which depends on the active resolve mode.

        Args:
            path (Path):
                The path to resolve, can be absolute or relative.

        Returns:
            Path:
                An absolute path resolved given the current state.

        Examples:
            - If the path is relative, it will be resolved relative to the current parent:
            ```python
            >>> from hydrolib.core.base.file_manager import FilePathResolver, ResolveRelativeMode
            >>> from pathlib import Path
            >>> resolver = FilePathResolver()
            >>> resolver.push_new_parent(Path("C:/project"), ResolveRelativeMode.ToParent)
            >>> resolver.resolve(Path("data/file.txt"))  # doctest: +SKIP
            Path('C:/project/data/file.txt')

            ```
            - If the path is absolute, it will be returned as is:
            ```python
            >>> resolver.resolve(Path("D:/absolute/path.txt"))  # doctest: +SKIP
            Path('D:/absolute/path.txt')

            ```
        """
        if path.is_absolute():
            return path

        parent = self.get_current_parent()
        return (parent / path).resolve()

    def push_new_parent(
        self, parent_path: Path, relative_mode: ResolveRelativeMode
    ) -> None:
        """Push a new parent_path with the given relative_mode to this FilePathResolver.

        Adds a new parent path to the resolver's stack. The parent_path is first resolved
        to an absolute path using the current state of the resolver. If the relative_mode
        is ToAnchor, the resolved path is also added to the anchors stack.

        Relative paths added to this FilePathResolver will be resolved with respect
        to the current state, i.e. similar to FilePathResolver.resolve.

        Args:
            parent_path (Path):
                The parent path to add, can be absolute or relative.
            relative_mode (ResolveRelativeMode):
                The relative mode used to resolve paths.
                ResolveRelativeMode.ToParent: Resolve relative to this parent.
                ResolveRelativeMode.ToAnchor: Resolve relative to this parent as an anchor.

        Examples:
            - Mode ToParent:
            ```python
            >>> from hydrolib.core.base.file_manager import FilePathResolver, ResolveRelativeMode
            >>> from pathlib import Path
            >>> resolver = FilePathResolver()
            >>> resolver.push_new_parent(Path("C:/project"), ResolveRelativeMode.ToParent)
            >>> print(resolver._parents) # doctest: +SKIP
            [(WindowsPath('C:/project'), <ResolveRelativeMode.ToParent: 0>)]
            >>> resolver.push_new_parent(Path("models"), ResolveRelativeMode.ToParent)
            >>> print(resolver._parents) # doctest: +SKIP
            >>> resolver.resolve(Path("model1.txt")) # doctest: +SKIP
            Path('C:/project/models/model1.txt')

            ```
            - Mode ToAnchor:
            ```python
            >>> resolver = FilePathResolver()
            >>> resolver.push_new_parent(Path("C:/project"), ResolveRelativeMode.ToParent)
            >>> print(resolver._parents) # doctest: +SKIP
            [(WindowsPath('C:/project'), <ResolveRelativeMode.ToParent: 0>)]
            >>> resolver.push_new_parent(Path("anchor"), ResolveRelativeMode.ToAnchor)
            >>> print(resolver._anchors) # doctest: +SKIP
            [WindowsPath('C:/project/anchor')]
            >>> print(resolver._parents) # doctest: +SKIP
            [
                (WindowsPath('C:/project'), <ResolveRelativeMode.ToParent: 0>),
                (WindowsPath('C:/project/anchor'), <ResolveRelativeMode.ToAnchor: 1>)
            ]
            >>> resolver.push_new_parent(Path("subdir"), ResolveRelativeMode.ToParent)
            >>> print(resolver._parents) # doctest: +SKIP
            [
                (WindowsPath('C:/project'), <ResolveRelativeMode.ToParent: 0>),
                (WindowsPath('C:/project/anchor'), <ResolveRelativeMode.ToAnchor: 1>),
                (WindowsPath('C:/project/anchor/subdir'), <ResolveRelativeMode.ToParent: 0>)
            ]
            >>> resolver.resolve(Path("file.txt")) # doctest: +SKIP
            Path('C:/project/anchor/file.txt')

            ```
        """
        absolute_parent_path = self.resolve(parent_path)
        if relative_mode == ResolveRelativeMode.ToAnchor:
            self._anchors.append(absolute_parent_path)

        self._parents.append((absolute_parent_path, relative_mode))

    def pop_last_parent(self) -> None:
        """Pop the last added parent from this FilePathResolver.

        Removes the most recently added parent path from the resolver's stack.
        If the parent had a relative_mode of ToAnchor, also removes the corresponding
        anchor from the anchors stack. If there are currently no parents defined,
        nothing will happen.

        Examples:
            ```python
            >>> resolver = FilePathResolver()
            >>> resolver.push_new_parent(Path("C:/project"), ResolveRelativeMode.ToParent)
            >>> print(resolver._parents) # doctest: +SKIP
            [(WindowsPath('C:/project'), <ResolveRelativeMode.ToParent: 0>)]
            >>> resolver.push_new_parent(Path("models"), ResolveRelativeMode.ToParent)
            >>> print(resolver._parents) # doctest: +SKIP
            [
                (WindowsPath('C:/project'), <ResolveRelativeMode.ToParent: 0>),
                (WindowsPath('C:/project/models'), <ResolveRelativeMode.ToParent: 0>)
            ]
            >>> resolver.pop_last_parent()
            >>> print(resolver._parents) # doctest: +SKIP
            [(WindowsPath('C:/project'), <ResolveRelativeMode.ToParent: 0>)]
            >>> resolver.resolve(Path("data.txt")) # doctest: +SKIP
            Path('C:/project/data.txt')

            ```

        Raises:
            No exceptions are raised, even if the parents stack is empty.
        """
        if not self._parents:
            return

        _, relative_mode = self._parents.pop()

        if relative_mode == ResolveRelativeMode.ToAnchor:
            self._anchors.pop()


class PathStyleValidator:
    """Class to take care of path style validation."""

    _os_path_style = get_path_style_for_current_operating_system()

    def validate(self, path_style: Optional[str]) -> PathStyle:
        """Validate.

        Validates the path style as string on whether it is a supported path style.
        If it is a valid path style the path style enum value will be return as a result.

        Args:
            path_style (Optional[str]): The path style as string value.

        Returns:
            PathStyle: The converted PathStyle object.

        Raises:
            ValueError: When an unsupported path style is passed.
        """
        if path_style is None:
            return self._os_path_style

        supported_path_styles = list(PathStyle)
        if path_style in supported_path_styles:
            return PathStyle(path_style)

        supported_path_style_str = ", ".join(([x.value for x in supported_path_styles]))
        raise ValueError(
            f"Path style '{path_style}' not supported. Supported path styles: {supported_path_style_str}"
        )


class ModelLoadSettings:
    """A class that holds the global settings for model loading."""

    def __init__(
        self, recurse: bool, resolve_casing: bool, path_style: PathStyle
    ) -> None:
        """Initializes a new instance of the ModelLoadSettings class.

        Args:
            recurse (bool): Whether or not to recursively load the whole model.
            resolve_casing (bool): Whether or not to resolve the file casing.
            path_style (PathStyle): Which path style is used in the loaded files.
        """
        self._recurse = recurse
        self._resolve_casing = resolve_casing
        self._path_style = path_style

    @property
    def recurse(self) -> bool:
        """Gets the recurse setting.

        Returns:
            bool: Whether or not to recursively load the whole model.
        """
        return self._recurse

    @property
    def resolve_casing(self) -> bool:
        """Gets the resolve casing setting.

        Returns:
            bool: Whether or not to resolve the file casing.
        """
        return self._resolve_casing

    @property
    def path_style(self) -> PathStyle:
        """Gets the path style setting.

        Returns:
            PathStyle: Which path style is used in the loaded files.
        """
        return self._path_style


class CachedFileModel:
    """CachedFileModel provides a simple structure to keep the Filemodel and checksum together."""

    _model: "FileModel"
    _checksum: str

    @property
    def model(self) -> "FileModel":
        """FileModel."""
        return self._model

    @property
    def checksum(self) -> str:
        """Checksum of the file the filemodel is based on."""
        return self._checksum

    def __init__(self, model: "FileModel", checksum: str) -> None:
        """Create a new empty CachedFileModel.

        Args:
            model (FileModel): filemodel to cache.
            checksum (str): checksum of the file.
        """
        self._model = model
        self._checksum = checksum


class FileModelCache:
    """FileModelCache provides a simple structure to register and retrieve FileModel objects."""

    def __init__(self):
        """Create a new empty FileModelCache."""
        self._cache_dict: Dict[Path, CachedFileModel] = {}

    def retrieve_model(self, path: Path) -> Optional["FileModel"]:
        """Retrieve model.

        Retrieve the model associated with the (absolute) path if
        it has been registered before, otherwise return None.

        Returns:
            [Optional[FileModel]]:
                The FileModel associated with the Path if it has been registered
                before, otherwise None.
        """
        cached_file_model = self._cache_dict.get(path, None)
        if cached_file_model is None:
            return None
        return cached_file_model.model

    def register_model(self, path: Path, model: "FileModel") -> None:
        """Register the model with the specified path in this FileModelCache.

        Args:
            path (Path): The path to associate the model with.
            model (FileModel): The model to be associated with the path.
        """
        checksum = self._get_checksum(path)
        self._cache_dict[path] = CachedFileModel(model, checksum)

    def is_empty(self) -> bool:
        """Whether or not this file model cache is empty.

        Returns:
            bool: Whether or not the cache is empty.
        """
        return not any(self._cache_dict)

    def _exists(self, path: Path) -> bool:
        """Whether or not the filepath is in the cache.

        Args:
            path (Path): The path to verify if it is already added in the cache.

        Returns:
            bool: Whether or not the path is in the cache.
        """
        return path in self._cache_dict

    def has_changed(self, path: Path) -> bool:
        """Whether or not the file in the filepath has changed from the cache.

        Args:
            path (Path): The path to verify verify against.

        Returns:
            bool: Whether or the file has changed.
            True when the file has changed.
            True when the file does not exist in caching reference.
            False when the file has not changed.
        """
        if not self._exists(path):
            return True

        checksum = self._get_checksum(path)
        return checksum != self._cache_dict.get(path).checksum

    def _get_checksum(self, path: Path) -> Optional[str]:
        return FileChecksumCalculator.calculate_checksum(path)


class FileCasingResolver:
    """Class for resolving file path in a case-insensitive manner."""

    def resolve(self, path: Path) -> Path:
        """Resolve the casing of a file path when the file does exist but not with the exact casing.

        Args:
            path (Path): The path of the file or directory for which the casing needs to be resolved.

        Returns:
            Path: The file path with the matched casing if a match exists; otherwise, the original file path.

        Raises:
            NotImplementedError: When this function is called with an operating system other than Windows, Linux or MacOS.
        """
        operating_system = get_operating_system()
        if operating_system == OperatingSystem.WINDOWS:
            return self._resolve_casing_windows(path)
        elif operating_system == OperatingSystem.LINUX:
            return self._resolve_casing_linux(path)
        elif operating_system == OperatingSystem.MACOS:
            return self._resolve_casing_macos(path)
        else:
            raise NotImplementedError(
                f"Path case resolving for operating system {operating_system} is not supported yet."
            )

    def _resolve_casing_windows(self, path: Path):
        return path

    def _resolve_casing_linux(self, path: Path):
        if path.exists():
            return path

        if not path.parent.exists() and not str_is_empty_or_none(path.parent.name):
            path = self._resolve_casing_linux(path.parent) / path.name

        return self._find_match(path)

    def _resolve_casing_macos(self, path: Path):
        if not str_is_empty_or_none(path.parent.name):
            path = self._resolve_casing_macos(path.parent) / path.name

        return self._find_match(path)

    def _find_match(self, path: Path):
        if path.parent.exists():
            for item in path.parent.iterdir():
                if item.name.lower() == path.name.lower():
                    return path.with_name(item.name)

        return path


class FileLoadContext:
    """FileLoadContext.

    FileLoadContext provides the context necessary to resolve paths
    during the init of a FileModel, as well as ensure the relevant models
    are only read once.
    """

    def __init__(self) -> None:
        """Create a new empty FileLoadContext."""
        self._path_resolver = FilePathResolver()
        self._cache = FileModelCache()
        self._file_casing_resolver = FileCasingResolver()
        self._file_path_style_converter = FilePathStyleConverter()
        self._load_settings: Optional[ModelLoadSettings] = None

    def initialize_load_settings(
        self, recurse: bool, resolve_casing: bool, path_style: PathStyle
    ):
        """Initialize the global model load setting. Can only be set once.

        Args:
            recurse (bool): Whether or not to recursively load the whole model.
            resolve_casing (bool): Whether or not to resolve the file casing.
            path_style (PathStyle): Which path style is used in the loaded files.
        """
        if self._load_settings is None:
            self._load_settings = ModelLoadSettings(recurse, resolve_casing, path_style)

    @property
    def load_settings(self) -> ModelLoadSettings:
        """Gets the model load settings.

        Raises:
            ValueError: When the model load settings have not been initialized yet.
        Returns:
            ModelLoadSettings: The model load settings.

        """
        if self._load_settings is None:
            raise ValueError(
                f"The model load settings have not been initialized yet. Make sure to call `{self.initialize_load_settings.__name__}` first."
            )

        return self._load_settings

    def retrieve_model(self, path: Optional[Path]) -> Optional["FileModel"]:
        """Retrieve the model associated with the path.

        If no model has been associated with the provided path, or the path is None,
        then None will be returned. Relative paths will be resolved based on the
        current state of the FileLoadContext.

        Returns:
            [Optional[FileModel]]:
                The file model associated with the provided path if it has been
                registered, else None.
        """
        if path is None:
            return None

        absolute_path = self._path_resolver.resolve(path)
        return self._cache.retrieve_model(absolute_path)

    def register_model(self, path: Path, model: "FileModel") -> None:
        """Associate the provided model with the provided path.

        Relative paths will be resolved based on the current state of the
        FileLoadContext.

        Args:
            path (Path): The relative path from which the model was loaded.
            model (FileModel): The loaded model.
        """
        absolute_path = self._path_resolver.resolve(path)
        self._cache.register_model(absolute_path, model)

    def cache_is_empty(self) -> bool:
        """Whether or not the file model cache is empty.

        Returns:
            bool: Whether or not the file model cache is empty.
        """
        return self._cache.is_empty()

    def get_current_parent(self) -> Path:
        """Get the current absolute path with which files are resolved.

        If the current mode is relative to the parent, the latest added
        parent is added. If the current mode is relative to an anchor
        path, the latest added anchor path is returned.

        Returns:
            Path:
                The absolute path to the current parent.
        """
        return self._path_resolver.get_current_parent()

    def resolve(self, path: Path) -> Path:
        """Resolve the provided path.

        If path is already absolute, it will be returned as is. Otherwise
        it will be resolved based on the current state of this FileLoadContext.

        Args:
            path (Path): The path to be resolved.

        Returns:
            Path: An absolute path resolved based on the current state.
        """
        return self._path_resolver.resolve(path)

    def push_new_parent(
        self, parent_path: Path, relative_mode: ResolveRelativeMode
    ) -> None:
        """Push a new parent_path with the given relative_mode on this FileLoadContext.

        Args:
            parent_path (Path): The parent path to be added to this FileLoadContext.
            relative_mode (ResolveRelativeMode): The relative mode.
        """
        self._path_resolver.push_new_parent(parent_path, relative_mode)

    def pop_last_parent(self) -> None:
        """Pop the last added parent off this FileLoadContext."""
        self._path_resolver.pop_last_parent()

    def resolve_casing(self, file_path: Path) -> Path:
        """Resolve the file casing for the provided file path.

        Args:
            file_path (Path): The file path to resolve the casing for.

        Returns:
            Path: The resolved file path.
        """
        if self.load_settings.resolve_casing:
            return self._file_casing_resolver.resolve(file_path)
        return file_path

    def convert_path_style(self, file_path: Path) -> Path:
        """convert_path_style.

        Resolve the file path by converting it from its own file path style to the path style for the current operating system.

        Args:
            file_path (Path): The file path to convert to the OS path style.

        Returns:
            Path: The resolved file path.
        """

        if file_path.is_absolute():
            return file_path

        converted_file_path = self._file_path_style_converter.convert_to_os_style(
            file_path, self.load_settings.path_style
        )
        return Path(converted_file_path)

    def is_content_changed(self, path: Path) -> bool:
        """Verify if the path is already known and if the content have changed.

        Relative paths will be resolved based on the current state of the
        FileLoadContext.

        Args:
            path (Path): The relative path from which the model was loaded.
            model (FileModel): The loaded model.

        Returns:
            True when the content from the path on the location has been changed.
            False when the content from the path is not priorly cached.
            False when the content from the path on the location has not been changed.
        """
        absolute_path = self._path_resolver.resolve(path)
        return self._cache.has_changed(absolute_path)


@contextmanager
def file_load_context() -> Generator[FileLoadContext, None, None]:
    """file_load_context.

    Provide a FileLoadingContext. If none has been created in the context of
    this call stack yet, a new one will be created, which will be maintained
    until it goes out of scope.

    Yields:
        [FileLoadContext]: The file load context.
    """
    file_loading_context = context_file_loading.get(None)
    context_reset_token = None

    if not file_loading_context:
        file_loading_context = FileLoadContext()
        context_reset_token = context_file_loading.set(file_loading_context)

    try:
        yield file_loading_context
    finally:
        if context_reset_token is not None:
            context_file_loading.reset(context_reset_token)


path_style_validator = PathStyleValidator()


def resolve_relative_to_root(file_path: Path, root_dir: Path) -> Path:
    """
    Resolve a file path relative to a root directory, handling both relative
    and absolute input paths robustly.

    This function ensures that a given `filepath` is correctly resolved under
    the `root_dir`. It supports cases where the `filepath` is:

    1. Already relative to `root_dir`.
    2. An absolute path or a longer relative path that includes `root_dir`.
    3. Outside the `root_dir`, in which case the path is returned as-is.

    This is useful in configuration handling or model loading workflows
    where file paths may come from user input, config files, or internal logic.

    Args:
        file_path (Path):
            The input path to a file. Can be relative, absolute, or already resolved.
        root_dir (Path):
            The base directory to which `filepath` should be made relative, if applicable.

    Returns:
        Path: An absolute path that is either:
            - Resolved under `root_dir` if possible, or
            - Left as-is and resolved if it lies outside `root_dir`.

    Raises:
        None


    Examples:
        ```python
        >>> root = Path("/models/config")
        ```
         - Case 1: file_path is a simple relative filename
         ```python
        >>> resolve_relative_to_root(Path("input.txt"), root) # doctest: +SKIP
        PosixPath('/models/config/input.txt')
        ```
        - Case 2: file_path is an absolute path inside root_dir
        ```python
        >>> resolve_relative_to_root(Path("/models/config/input.txt"), root)
        PosixPath('/models/config/input.txt')
        ```
        - Case 3: file_path is an absolute path outside root_dir
        ```python
        >>> resolve_relative_to_root(Path("/other/data/input.txt"), root) # doctest: +SKIP
        PosixPath('/other/data/input.txt')
        ```
        - not includes:
        - Case 2: file_path is a longer relative path starting with root_dir
        ```python
        >>> resolve_relative_to_root(Path("models/config/input.txt"), Path("models/config"))
        PosixPath('/models/config/input.txt')
        ```
        - or
        ```python
        >>> resolve_relative_to_root(Path('data/input/root/input.txt'), Path('C:/abs/path/data/input/root')) # doctest: +SKIP
        WindowsPath('C:/abs/path/data/input/root/input.txt')
        # but given
        WindowsPath('C:/abs/path/data/input/root/data/input/root/input.txt')
        ```

    Notes:
        - This function **does not use the current working directory** to resolve paths.
        - It is designed for use cases where `root_dir` is the only authoritative base.
        - Symbolic links will be resolved (because `.resolve()` is used).

    """
    root_dir = root_dir.resolve()
    if not file_path.is_absolute():
        # Always resolve relative paths *relative to root_dir*
        file_path_resolved = (root_dir / file_path).resolve()
    else:
        file_path_resolved = file_path.resolve()

    try:
        relative_path = file_path_resolved.relative_to(root_dir)
        final_path = root_dir / relative_path
    except ValueError:
        final_path = file_path_resolved

    return final_path
